These are just code fragments, not complete programs or routines.  I have
not given layouts of things like "itrb" (image transform request block)
because they're application-defined and not important to the calling
sequence.  Instead, pay attention to "csb" (codec state block) fields.

This file is best viewed with tab spacing set to 4


//
//
//
//	G4-encode
//
//
//

void G4encode(ITRB &itrb)
{
	CSB
		csb;
	unsigned long
		length;

	memset(&csb, 0, sizeof csb);

	// Destination for compressed output
	csb.bs.rawdata = csb.bs.rawcp = (uchar *)itrb.bits;
	csb.bs.rawcc = 0;

	// Image size
	csb.bs.rowpixels = itrb.width;
	csb.bs.rows 	 = itrb.height;
	csb.bs.rowbytes  = (itrb.width + 7) / 8;

	// Vertical resolution
	csb.bs.yresolution = itrb.resY *
		itrb.resUnit == 1 ? 0.254 : 0.1;

	// Encoding options
	csb.bs.groupoptions = TIFF_Group3Options::G3_2DENCODING;

	// Reference line
	csb.es.refline = new uchar[csb.bs.rowbytes];

	// Uncompressed image length.  Bits per component
	// is known to be 1.
	length = (itrb.width * itrb.height + 7) / 8;

	// Do G4 compression
	Fax4Encode(&csb, (uchar *)itrb.raster, length);

	// Record length of compressed image
	itrb.cilength = csb.bs.rawcc;

	// Clean up
	delete [] csb.es.refline;
}


//
//
//
// G4-decode
//
//
//


//
//
// Method:         Image()
//
// Class:			Image()
//
// Purpose:		Constructor
//
// Inputs:         Image File Directory (IFD), component of a TIFF object
//
// Outputs:        None
//
// Returns:        Constructed, decompressed image
//
// Side-effects:   None
//
// Notes:			Minimum granularity of the decompressed image is one
//					byte per pixel.
//
//

Image::Image(Ifd &ifd)
{
	Field
		*f;
	uint16
		compression;

	raster = NULL;
	compression = ifd.field(TIFF_Tag::COMPRESSION)->valueOffset;

	// Every IFD must have these.

	cx = ifd.field(TIFF_Tag::IMAGEWIDTH)->valueOffset;
	cy = ifd.field(TIFF_Tag::IMAGELENGTH)->valueOffset;

	// Every IFD should also have these but some TIFF authors omit them
	// for bitonal images.

	f = ifd.field(TIFF_Tag::BITSPERSAMPLE);
	if (f)
		bpc = f->valueOffset;
	else if (compression == TIFF_Compression::CCITTFAX3
		 ||  compression == TIFF_Compression::CCITTFAX4)
		bpc = 1;

	f = ifd.field(TIFF_Tag::ORIENTATION);
	if (f)
		orientation = f->valueOffset;
	else
		orientation = 1;

	f = ifd.field(TIFF_Tag::SAMPLESPERPIXEL);
	if (f)
		components = f->valueOffset;
	else
		components = 1;

	// Now I have enough information to allocate the raster storage.
	//
	// Note that I ignore bits-per-sample and allocate one byte per
	// pixel component.

	raster = new unsigned char[cx * cy * components];
	if (raster == NULL)
		return;					// Should I throw an exception?

	// Now it is time to invoke the appropriate decompressor

	switch (compression)
	{
	case TIFF_Compression::CCITTFAX3:
	case TIFF_Compression::CCITTFAX4:
		g4decode(ifd);
//		crop();
		break;
	case TIFF_Compression::JPEG:
	case TIFF_Compression::OJPEG:
		jpegdecode(ifd);
		break;
	default:
		delete [] raster;
		raster = NULL;			// Should I throw an exception?
	}
}

Image::~Image()
{
	if (raster)
		delete [] raster;
}


Image &Image::g4decode(Ifd &ifd)
{
	Field
		*f;
	CSB
		csb;
	int32
		*pi,
		bytes;
	unsigned char
		pixelValue[2],
		*bits;

	csb.bs.groupoptions = TIFF_Group3Options::G3_2DENCODING;

	// Store the vertical resolution in the codec state block.

	f = ifd.field(TIFF_Tag::YRESOLUTION);
	pi = (int32 *)f->pValue;
	csb.bs.yresolution = (float)pi[0] / (float)pi[1];
	f = ifd.field(TIFF_Tag::RESOLUTIONUNIT);
	csb.bs.resolutionunit = f->valueOffset;

	// Image dimensions

	csb.bs.rowpixels = cx;
	csb.bs.rows      = cy;

	if (bpc != 1)
	{
		cout << "Bits per sample is " << bpc
			 << ".  For a CCITT G4 image it should be 1" << endl;
		return *this;		// +++ throw an exception?  Probably yes.
	}

	// Row granularity.

	csb.bs.rowbytes = roundup(csb.bs.rowpixels, 8) / 8;

	{							// 2d decoding
		uint32 nruns = 2 * roundup(csb.bs.rowpixels, 8);

		csb.ds.runs = new uint16[nruns];
		if (csb.ds.runs == NULL)
		{
			cout << "No space for Group 3/4 run arrays" << endl;
			return *this;		// +++ throw exception
		}
		csb.ds.curruns = csb.ds.runs;
		csb.ds.refruns = csb.ds.runs + (nruns >> 1);
	}

	f = ifd.field(TIFF_Tag::PHOTOMETRIC);
	if (f)
		switch (f->valueOffset)
		{
		case TIFF_Photometric::MINISBLACK:
			pixelValue[0] = 0x00;
			pixelValue[1] = 0xFF;
			break;
		case TIFF_Photometric::MINISWHITE:
		default:
			pixelValue[0] = 0xFF;
			pixelValue[1] = 0x00;
		}
	else
	{
		pixelValue[0] = 0xFF;
		pixelValue[1] = 0x00;
	}

	// Set up the codec output pointer

	bytes = csb.bs.rows * csb.bs.rowbytes;
	bits = new unsigned char [bytes];
	if (bits == NULL)
		return *this;		// +++ throw exception

	// Set up the codec input array and length.

	{
		unsigned char
			**ppuch;
		unsigned long
			*pul;
		f = ifd.field(TIFF_Tag::STRIPOFFSETS);
		ppuch = (unsigned char **)f->pValue;
		csb.bs.rawdata = csb.bs.rawcp = *ppuch;
		f = ifd.field(TIFF_Tag::STRIPBYTECOUNTS);
		pul = (unsigned long *)f->pValue;
		csb.bs.rawcc = *pul;
	}

	// Now decode the image into the bitmap

	Fax3PreDecode(&csb, TIFF_FillOrder::MSB2LSB);
	Fax4Decode(&csb, bits, bytes);

	// Image should now be decoded, one bit per pixel.  Now I expand it
	// to one byte per pixel.  (Makes subsequent operations a lot easier.)

	f = ifd.field(TIFF_Tag::ORIENTATION);
	if (f)
		orientation = f->valueOffset;
	else
		orientation = 1;

	// PM orientation is left to right, bottom to top which corresponds to
	// a TIFF tag of 4.  While it would be most convenient for me to use
	// that orientation (simplifying the construction of bitmaps) in some
	// respects it is more appropriate to use orientation 1.  The main
	// reason for that is that the JPEG decoder will use that orientation
	// by default.
	//
	// Orientations 5 to 8 imply a 90% rotation in some sense.  I do not
	// have to worry about those.  The orientations I get are in the range
	// 1 to 4.


